@@ -37,6 +37,7 @@
 #include "vino-util.h"
 #include "vino-enums.h"
 #include "vino-background.h"
+#include "vino-upnp.h"
 #include <sys/poll.h>
 #include <dbus/dbus-glib.h>
 #include <gtk/gtk.h>
@@ -60,6 +61,7 @@ struct _VinoServerPrivate
   VinoPrompt       *prompt;
   VinoStatusIcon   *icon;
   VinoDBusListener *listener;
+  VinoUpnp         *upnp;
 
   GIOChannel       *io_channel[RFB_MAX_SOCKETLISTEN];
   guint             io_watch[RFB_MAX_SOCKETLISTEN];
@@ -84,6 +86,7 @@ struct _VinoServerPrivate
   guint             use_alternative_port : 1;
   guint             lock_screen : 1;
   guint             disable_background : 1;
+  guint             use_upnp : 1;
 };
 
 struct _VinoClient
@@ -114,7 +117,8 @@ enum
   PROP_VNC_PASSWORD,
   PROP_PORT,
   PROP_LOCK_SCREEN,
-  PROP_DISABLE_BACKGROUND
+  PROP_DISABLE_BACKGROUND,
+  PROP_USE_UPNP
 };
 
 static enum rfbNewClientAction vino_server_auth_client (VinoServer *server,
@@ -128,13 +132,14 @@ static void vino_server_update_security_types (VinoServer *server);
 
 static gpointer parent_class;
 
-static void
-vino_server_lock_screen (VinoServer *server)
-{
 #define GNOME_SCREENSAVER_BUS_NAME  "org.gnome.ScreenSaver"
 #define GNOME_SCREENSAVER_INTERFACE "org.gnome.ScreenSaver"
 #define GNOME_SCREENSAVER_PATH      "/org/gnome/ScreenSaver"
 
+static void
+vino_server_lock_screen (VinoServer *server)
+{
+
   DBusGConnection *connection;
   GError          *error;
   DBusGProxy      *proxy;
@@ -163,11 +168,42 @@ vino_server_lock_screen (VinoServer *server)
 
   g_object_unref (proxy);
   dbus_g_connection_unref (connection);
+}
+
+static void
+vino_server_unlock_screen (void)
+{
+
+  DBusGConnection *connection;
+  GError          *error;
+  DBusGProxy      *proxy;
+
+  dprintf(DBUS, "Unlocking screen via gnome-screensaver\n");
+
+  error = NULL;
+  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
+  if (!connection)
+    {
+      g_printerr (_("Failed to open connection to bus: %s\n"),
+                  error->message);
+      g_error_free (error);
+      return;
+    }
+
+  proxy = dbus_g_proxy_new_for_name (connection,
+                                     GNOME_SCREENSAVER_BUS_NAME,
+                                     GNOME_SCREENSAVER_PATH,
+                                     GNOME_SCREENSAVER_INTERFACE);
+
+  dbus_g_proxy_call_no_reply (proxy, "SimulateUserActivity", G_TYPE_INVALID);
+
+  g_object_unref (proxy);
+  dbus_g_connection_unref (connection);
+}
 
 #undef GNOME_SCREENSAVER_BUS_NAME
 #undef GNOME_SCREENSAVER_INTERFACE
 #undef GNOME_SCREENSAVER_PATH
-}
 
 void
 vino_server_set_disable_background (VinoServer *server,
@@ -193,12 +229,55 @@ vino_server_get_disable_background (VinoServer *server)
   return server->priv->disable_background;
 }
 
+static void
+vino_server_control_upnp (VinoServer *server)
+{
+  if (server->priv->use_upnp && !server->priv->on_hold)
+    {
+      if (!server->priv->upnp)
+	server->priv->upnp = vino_upnp_new ();
+      vino_upnp_add_port (server->priv->upnp, server->priv->rfb_screen->rfbPort);
+    }
+  else
+    if (server->priv->upnp)
+      {
+	g_object_unref (server->priv->upnp);
+	server->priv->upnp = NULL;
+      }
+}
+
+void
+vino_server_set_use_upnp (VinoServer *server,
+                          gboolean use_upnp)
+{
+  g_return_if_fail (VINO_IS_SERVER (server));
+
+  use_upnp = use_upnp != FALSE;
+
+  if (server->priv->use_upnp != use_upnp)
+    {
+      server->priv->use_upnp = use_upnp;
+      vino_server_control_upnp (server);
+      g_object_notify (G_OBJECT (server), "use-upnp");
+    }
+}
+
+gboolean
+vino_server_get_use_upnp (VinoServer *server)
+{
+  g_return_val_if_fail (VINO_IS_SERVER (server), FALSE);
+
+  return server->priv->use_upnp;
+}
+
 static void
 vino_server_client_accepted (VinoServer *server,
                              VinoClient *client)
 {
   vino_status_icon_add_client (server->priv->icon, client);
 
+  vino_server_unlock_screen ();
+
   if (vino_server_get_disable_background (server))
      vino_background_draw (FALSE);
 }
@@ -813,6 +892,9 @@ vino_server_clipboard_cb (GtkClipboard *cb,
   gchar *text, *out;
   size_t a, b;
 
+  if (!server->priv->clients)
+    return;
+
   if (server->priv->on_hold)
     return;
 
@@ -833,7 +915,6 @@ vino_server_clipboard_cb (GtkClipboard *cb,
     }
 
   g_free (text);
-
 }
 
 static void
@@ -947,6 +1028,7 @@ vino_server_init_from_screen (VinoServer *server,
                     server);
 
   server->priv->icon = vino_status_icon_new (server, server->priv->screen);
+  server->priv->upnp = NULL;
 }
 
 static void
@@ -1012,6 +1094,10 @@ vino_server_finalize (GObject *object)
   if (server->priv->icon)
     g_object_unref (server->priv->icon);
   server->priv->icon = NULL;
+
+  if (server->priv->upnp)
+    g_object_unref (server->priv->upnp);
+  server->priv->upnp = NULL;
   
   g_free (server->priv);
   server->priv = NULL;
@@ -1066,6 +1152,9 @@ vino_server_set_property (GObject      *object,
     case PROP_DISABLE_BACKGROUND:
       vino_server_set_disable_background (server, g_value_get_boolean (value));
       break;
+    case PROP_USE_UPNP:
+      vino_server_set_use_upnp (server, g_value_get_boolean (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1121,6 +1210,9 @@ vino_server_get_property (GObject    *object,
     case PROP_DISABLE_BACKGROUND:
       g_value_set_boolean (value, server->priv->disable_background);
       break;
+    case PROP_USE_UPNP:
+      g_value_set_boolean (value, server->priv->use_upnp);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1305,6 +1397,18 @@ vino_server_class_init (VinoServerClass *klass)
                                                          G_PARAM_STATIC_NICK |
                                                          G_PARAM_STATIC_BLURB));
 
+  g_object_class_install_property (gobject_class,
+				   PROP_USE_UPNP,
+				   g_param_spec_boolean ("use-upnp",
+							 "Use UPNP",
+							 "Whether to use UPNP",
+							 FALSE,
+                                                         G_PARAM_READWRITE   |
+                                                         G_PARAM_CONSTRUCT   |
+                                                         G_PARAM_STATIC_NAME |
+                                                         G_PARAM_STATIC_NICK |
+                                                         G_PARAM_STATIC_BLURB));
+
 }
 
 GType
@@ -1436,6 +1540,8 @@ vino_server_set_use_alternative_port (VinoServer *server,
 
           rfbSetAutoPort (server->priv->rfb_screen,
                           !server->priv->use_alternative_port);
+
+	  vino_server_control_upnp (server);
         }
 
       g_object_notify (G_OBJECT (server), "use-alternative-port");
@@ -1462,7 +1568,10 @@ vino_server_set_alternative_port (VinoServer *server,
 
       if (server->priv->rfb_screen &&
           server->priv->use_alternative_port)
-        rfbSetPort (server->priv->rfb_screen, server->priv->alternative_port);
+	{
+	  rfbSetPort (server->priv->rfb_screen, server->priv->alternative_port);
+	  vino_server_control_upnp (server);
+	}
 
       g_object_notify (G_OBJECT (server), "alternative-port");
     }
@@ -1507,9 +1616,11 @@ vino_server_set_on_hold (VinoServer *server,
 	}
 
       g_object_notify (G_OBJECT (server), "on-hold");
-      
+
       if (server->priv->icon)
         vino_status_icon_update_state (server->priv->icon);
+
+      vino_server_control_upnp (server);
     }
 }
 
